/* 
*/

#include <Arduino.h>
#include <WiFiNINA.h>
#include "PubSubClient.h"
#include <avr/dtostrf.h>
#include <RTCZero.h>
#include "time.h"

RTCZero rtc;
const int GMT = -8; //change this to adapt it to your time zone

//                  ==========================  WIFI & Ubidots account info ======================

#define SSID_NAME "xxxxxx" // Put your WifiSSID here
#define SSID_PASS "xxxxxx" // Put your wifi password here

#define MQTT_CLIENT_NAME "nnnnnnnnnn" // MQTT client Name - enter your own 8-12 alphanumeric character ASCII string; 
#define TOKEN "BBFF-tttttttttttttttttttttttttttttt"  // Cat Door Test
#define DEVICE_LABEL "lllllll-llll-ll-llll

char mqttBroker[]  = "industrial.api.ubidots.com";
const int mqttPort = 1883;  // 8883 for secure

//WiFiSSLClient ubiWiFi;
WiFiClient WiFiCl;
PubSubClient ubiPub(WiFiCl);

//                   ============ local Constants & Variables ============

unsigned long TS_EPOCH = 0;
unsigned long TS_OFFSET = 0;

/********************************
 * Auxiliary Functions
 *******************************/

bool setRTC() {
  unsigned long epoch = 0;
  int tries = 0;
  while (tries < 6) {
    epoch = WiFi.getTime();
    if (epoch > 0)
      break;
    tries++;
    delay(1000);
  }
  if (epoch == 0) {
    Serial.println("NTP unreachable!!");
    return false;
  }
  else {
    Serial.print("Epoch received: "); Serial.println(epoch);
    rtc.setEpoch(epoch);
    return true;
  }
}

bool setRTCNB() {
  unsigned long epoch = WiFi.getTime();
  if (epoch > 0) {
    Serial.print("Epoch received: "); Serial.println(epoch);
    rtc.setEpoch(epoch);
    return true;
  } else {
    Serial.println("NTP unreachable!!");
    return false;
  }
}

void printWiFiSt(byte st) {
  switch (st) {
    case WL_CONNECTED:
      Serial.print("CONNECTED");
      break;
    case WL_AP_CONNECTED:
      Serial.print("AP CONNECTED");
      break;
    case WL_AP_LISTENING:
      Serial.print("AP LISTENING");
      break;
//    case WL_NO_SHIELD:
//      Serial.print("NO SHIELD");
//      break;
    case WL_NO_MODULE:
      Serial.print("NO MODULE");
      break;
    case WL_IDLE_STATUS:
      Serial.print("IDLE STATUS");
      break;
    case WL_NO_SSID_AVAIL:
      Serial.print("NO SSID AVAIL");
      break;
    case WL_SCAN_COMPLETED:
      Serial.print("SCAN COMPLETED");
      break;
    case WL_CONNECT_FAILED:
      Serial.print("CONNECT FAILED");
      break;
    case WL_CONNECTION_LOST:
      Serial.print("CONNECTION LOST");
      break;
    case WL_DISCONNECTED:
      Serial.print("DISCONNECTED");
      break;
  }
}

void printWiFiStatus(byte st) {
  Serial.print("WiFi Status: "); Serial.print(st);
  Serial.print(" - "); printWiFiSt(st);
  Serial.print(", SSID: "); Serial.print(WiFi.SSID());
  Serial.print(", IP Address: "); Serial.print(WiFi.localIP());
  Serial.print(", RSSI:"); Serial.print(WiFi.RSSI()); Serial.print(" dBm");
  Serial.println();
}

void printWiFiStatus() {
  byte st = WiFi.status();
  printWiFiStatus(st);
}

void ReconnectWiFi() {  // non-blocking...
  unsigned long sTime = millis();
  byte st = WiFi.status();
  printWiFiStatus(st);
  WiFi.end();
  delay(200);
  WiFi.begin(SSID_NAME, SSID_PASS);  // try again?
  delay(200);
  st = WiFi.status();
  printWiFiStatus(st);
  unsigned long rTime = millis()-sTime;
  Serial.print("==> ReconnectWiFi, elapsed = "); Serial.println(rTime);
//  ubiSend("wifi-reconnect-time",rTime/1000.0);
}

//                                                   ====================== UBIDOTS ========================

// UBIDOTS Callback
void CallBack(char* topic, byte* payload, unsigned int length) {
  char p[length + 1];
  memcpy(p, payload, length);
  p[length] = NULL;
  String message(p);
  int nMessage = message.toInt();
  char * pch;
  char * varname;
  int k = 0;

  Serial.print("Callback - topic: "); Serial.print(topic);
  pch = strtok(topic,"/");
  while (pch != NULL) {
    k++;
//    Serial.print(" "); Serial.print(k); Serial.print(":"); Serial.print(pch); Serial.print(", ");
    if (k == 4) {
      varname = pch;
    }
    pch = strtok(NULL, "/");
  }
  Serial.print(", payload: "); Serial.write(payload, length); 
  Serial.println();

  if (strcmp(varname,"btn-open-1") == 0) {
    if (nMessage) {
      ubiSend(varname,0,true);  // reset the button 
    }
  }
  else if (strcmp(varname,"btn-open-2") == 0) {
    if (nMessage) {
      ubiSend(varname,0,false);  // reset the button 
    }
  }

}

void ubiSubscribe(char varname[]) {
 char topicSubscribe[100];
  sprintf(topicSubscribe, "/v1.6/devices/%s/%s/lv", DEVICE_LABEL, varname);
  ubiPub.subscribe(topicSubscribe);
  Serial.print("Subscribe: "); Serial.println(topicSubscribe);
}

void setCallbacks() {
  ubiSubscribe("btn-open-1");
  ubiSubscribe("btn-open-2");
}

void ReconnectNB() {  // non-blocking
  unsigned long sTime = millis();
  Serial.print("MQTT connect: "); Serial.print(mqttBroker);
  Serial.print(", port: "); Serial.print(mqttPort); 
  Serial.print(", client: "); Serial.println(MQTT_CLIENT_NAME);
  // Attempt to connect
  if (ubiPub.connect(MQTT_CLIENT_NAME, TOKEN, "")) {
    Serial.println("MQTT connected, setting callbacks...");
    setCallbacks();  // must be done after reconnect
  } else {
    Serial.print("Failed, rc="); Serial.println(ubiPub.state());
  }
  unsigned long rTime = millis()-sTime;
  Serial.print("==> ReconnectNB, elapsed = "); Serial.println(rTime);
//  ubiSend("ubi-reconnect-time",rTime/1000.0);
}

// Publish Queue creation:
struct PublishData {
  unsigned long timestamp;
  unsigned long milliseconds;
  char varname[20];
  char value[10];
};

void ubiSend(char varname[], double var, bool ts) {
 char payload[200];
 char topic[150];
 char str_data[10];
  unsigned long tse = rtc.getEpoch();
  if (tse>TS_EPOCH) {  // has epoch ticked?
    TS_EPOCH = tse;  // get newest epoch
    TS_OFFSET = 0;  // reset offset
  }  else {
    TS_OFFSET += 200;  // increment ms to keep dots in sequence.
  }
  if (TS_OFFSET>999) {
    TS_EPOCH += 1;
    TS_OFFSET = 0;
  }
  sprintf(topic, "%s%s", "/v1.6/devices/", DEVICE_LABEL);
  sprintf(payload, "%s", ""); // Cleans the payload ???
  sprintf(payload, "{\"%s\":", varname); // Adds the variable label
  dtostrf(var, 3, 2, str_data);  // 2 is mininum width, 0 is precision
  sprintf(payload, "%s{\"value\":%s", payload, str_data); // Add the value
  if (ts) {
   sprintf(payload, "%s,\"timestamp\":%u%3.3u", payload, TS_EPOCH, TS_OFFSET); // Adds the timestamp
  }
  sprintf(payload, "%s}}", payload);  // add ending braces
  bool rc = ubiPub.publish(topic, payload);
  Serial.print("Sent: "); Serial.print(topic); Serial.print(","); Serial.print(payload);
  Serial.print(", len="); Serial.print(strlen(topic)+strlen(payload)); Serial.print(", rc="); Serial.println(rc);
}

void ubiSend(char varname[], double var) {
  ubiSend(varname, var, false);  
}

//                                                    ====================== SETUP ========================

void setup() {
  // Initialize serial and wait for port to open:
  Serial.begin(115200);
  // maybe wait for a Serial Monitor ...
  delay(1500); 

  // attempt to connect to WiFi network:
  Serial.print("Connecting to SSID: "); Serial.println(SSID_NAME);
  unsigned long sTime = millis();
  WiFi.begin(SSID_NAME, SSID_PASS);
  delay(200);
  printWiFiStatus();
  if (WiFi.status() == WL_CONNECTED) {
    unsigned long rTime = millis()-sTime;
    Serial.print("==> ConnectWiFi, elapsed = "); Serial.println(rTime);
//    ubiSend("wifi-reconnect-time",rTime/1000.0);
    setRTC();
    Serial.print("Connecting to "); Serial.println(mqttBroker);
    ubiPub.setServer(mqttBroker, mqttPort);
    ubiPub.setCallback(CallBack);
    ReconnectNB();
  }
//  printTime();

  Serial.println("Setup done");
}

// ====================== LOOP ========================

unsigned long reconTime = 0;
unsigned long wifiTime = 0;
unsigned long now;

void loop() {

  byte st = WiFi.status();
  if (st != WL_CONNECTED && st<10) {  // bug in WiFiNiNa !!!
    now = millis();
    if ((now-wifiTime)>2000) {  // attempt reconnect every 2 seconds
      ReconnectWiFi();
      wifiTime = now;
    }
  } else {
    if (rtc.getEpoch() == 0) {
      setRTCNB();
    }
    if (!ubiPub.connected()) {
      now = millis();
      if ((now-reconTime) > 2000) {  // attempt reconnect every 2 seconds
        ReconnectNB();
        reconTime = now;
      }
    } else {
      ubiPub.loop();
//      ubiPublishQ();  // publish the oldest item in the queue
    }
  }  
  
}
